from optparse import OptionParser
import sys
import os
import codecs
from pyinstrument import Profiler
from pyinstrument.profiler import get_recorder_class, get_renderer_class
from .six import exec_


def main():
    usage = ("usage: pyinstrument [options] scriptfile [arg] ...")
    parser = OptionParser(usage=usage)
    parser.allow_interspersed_args = False

    parser.add_option('', '--setprofile',
        dest='setprofile', action='store_true',
        help='run in setprofile mode, instead of signal mode', default=False)

    parser.add_option('', '--html',
        dest="output_html", action='store_true',
        help="output HTML instead of text", default=False)
    parser.add_option('', '--flame',
        dest='output_flame', action='store_true',
        help='output an HTML flame chart', default=False)
    parser.add_option('-r', '--renderer',
        dest='output_renderer', action='store', type='string',
        help='python import path to a renderer class', default=None)

    parser.add_option('-o', '--outfile',
        dest="outfile", action='store',
        help="save report to <outfile>", default=None)

    parser.add_option('', '--unicode',
        dest='unicode', action='store_true',
        help='force unicode text output')
    parser.add_option('', '--no-unicode',
        dest='unicode', action='store_false',
        help='force ascii text output')

    parser.add_option('', '--color',
        dest='color', action='store_true',
        help='force ansi color text output')
    parser.add_option('', '--no-color',
        dest='color', action='store_false',
        help='force no color text output')

    if not sys.argv[1:]:
        parser.print_help()
        sys.exit(2)

    (options, args) = parser.parse_args()
    sys.argv[:] = args

    if len(args) > 0:
        progname = args[0]
        sys.path.insert(0, os.path.dirname(progname))

        with open(progname, 'rb') as fp:
            code = compile(fp.read(), progname, 'exec')
        globs = {
            '__file__': progname,
            '__name__': '__main__',
            '__package__': None,
        }

        if options.output_renderer:
            renderer = options.output_renderer
        elif options.output_html:
            renderer = 'html'
        else:
            renderer = 'text'

        recorder = get_renderer_class(renderer).preferred_recorder

        profiler = Profiler(recorder=recorder)

        profiler.start()

        try:
            exec_(code, globs, None)
        except (SystemExit, KeyboardInterrupt):
            pass

        profiler.stop()

        if options.outfile:
            f = codecs.open(options.outfile, 'w', 'utf-8')
        else:
            f = sys.stdout

        renderer_kwargs = {}

        if renderer == 'text':
            unicode_override = options.unicode != None
            color_override = options.color != None
            unicode = options.unicode if unicode_override else file_supports_unicode(f)
            color = options.color if color_override else file_supports_color(f)
            
            renderer_kwargs = {'unicode': unicode, 'color': color}

        f.write(profiler.output(renderer=renderer, **renderer_kwargs))
        f.close()
    else:
        parser.print_usage()


def file_supports_color(file_obj):
    """
    Returns True if the running system's terminal supports color, and False
    otherwise.

    Borrowed from Django
    https://github.com/django/django/blob/master/django/core/management/color.py
    """
    plat = sys.platform
    supported_platform = plat != 'Pocket PC' and (plat != 'win32' or
                                                  'ANSICON' in os.environ)

    is_a_tty = hasattr(file_obj, 'isatty') and file_obj.isatty()
    if not supported_platform or not is_a_tty:
        return False
    return True


def file_supports_unicode(file_obj):
    encoding = getattr(file_obj, 'encoding', None)
    if not encoding:
        return False

    codec_info = codecs.lookup(encoding)

    if 'utf' in codec_info.name:
        return True
    return False


if __name__ == '__main__':
    main()



from collections import deque
import abc
from .frame import TimeAggregatingFrame, TimelineFrame
from . import six


class Recorder(object):
    '''
    An object that can record frames and return a tree structure of Frame objects
    '''

    def record_frame(self, frame, time):
        '''
        Record that `time` seconds was spent in `frame`. `frame` is an interpreter frame.
        '''
        raise NotImplementedError()

    def root_frame(self):
        '''
        Output a Frame object that represents the parsed tree of frames
        '''
        raise NotImplementedError()


class TimeAggregatingRecorder(Recorder):
    '''
    Records each frame according to the frame identifiers of the call stack. Does not preserve
    the order in which frames are visited.
    '''
    def __init__(self):
        self.stack_self_time = {}

    def record_frame(self, frame, time):
        stack = self._call_stack_for_frame(frame)
        self.stack_self_time[stack] = self.stack_self_time.get(stack, 0) + time

    def _call_stack_for_frame(self, frame):
        result_list = deque()

        while frame is not None:
            result_list.appendleft(self._identifier_for_frame(frame))
            frame = frame.f_back

        return tuple(result_list)

    def _identifier_for_frame(self, frame):
        # we use a string here as a tuple hashes slower and this is used as a key in a dictionary
        return '%s\x00%s\x00%i' % (
            frame.f_code.co_name, frame.f_code.co_filename, frame.f_code.co_firstlineno
        )

    def root_frame(self):
        root_frame = TimeAggregatingFrame()

        # define a recursive function that builds the hierarchy of frames given the
        # stack of frame identifiers
        def frame_for_stack(stack):
            if len(stack) == 0:
                return root_frame

            parent = frame_for_stack(stack[:-1])
            frame_name = stack[-1]

            if frame_name not in parent.children_dict:
                child = TimeAggregatingFrame(frame_name, parent)
                parent.add_child(child)

            return parent.children_dict[frame_name]

        for stack, self_time in self.stack_self_time.items():
            frame_for_stack(stack).self_time = self_time

        return root_frame


class TimelineRecorder(Recorder):
    '''
    Records every frame separately so that the order that frames are visited is preserved.
    '''
    def __init__(self):
        self.frames = []

    def record_frame(self, frame, time):
        stack = self._call_stack_for_frame(frame)
        self.frames.append((stack, time))

    def _call_stack_for_frame(self, frame):
        result_list = deque()

        while frame is not None:
            result_list.appendleft(self._identifier_for_frame(frame))
            frame = frame.f_back

        return tuple(result_list)

    def _identifier_for_frame(self, frame):
        # we use a string here as a tuple hashes slower and this is used as a key in a dictionary
        return '%s\x00%s\x00%i' % (
            frame.f_code.co_name, frame.f_code.co_filename, frame.f_code.co_firstlineno
        )

    def root_frame(self):
        root_frame = TimelineFrame()

        frame_objects = []

        for frame_tuple in self.frames:
            identifier_stack = frame_tuple[0]
            time = frame_tuple[1]

            # now we must create a stack of frame objects and assign this time to the leaf
            for stack_depth, frame_identifier in enumerate(identifier_stack):
                if stack_depth < len(frame_objects):
                    if frame_identifier != frame_objects[stack_depth].identifier:
                        # trim any frames after and including this one
                        del frame_objects[stack_depth:]

                if stack_depth >= len(frame_objects):
                    if stack_depth == 0:
                        parent = root_frame
                    else:
                        parent = frame_objects[stack_depth-1]

                    frame = TimelineFrame(frame_identifier, parent)
                    parent.add_child(frame)
                    frame_objects.append(frame)

            # trim any extra frames
            del frame_objects[stack_depth+1:]

            # assign the time to the final frame
            frame_objects[-1].self_time += time

        return root_frame



import importlib

def object_with_import_path(import_path):
    if '.' not in import_path:
        raise ValueError("Can't import '%s', it is not a valid import path" % import_path)
    module_path, object_name = import_path.rsplit('.', 1)

    module = importlib.import_module(module_path)
    return getattr(module, object_name)



# -*- coding: utf-8 -*-
import abc
import os
from . import six

class Renderer(object):
    preferred_recorder = ''

    def render(self, frame):
        ''' 
        Return a string that contains the rendered form of `frame`
        '''
        raise NotImplementedError()


class ConsoleRenderer(Renderer):
    preferred_recorder = 'time_aggregating'

    def __init__(self, unicode=False, color=False, **kwargs):
        self.unicode = unicode
        self.color = color
        super(ConsoleRenderer, self).__init__(**kwargs)

    def render(self, frame, indent=u'', child_indent=u''):
        colors = colors_enabled if self.color else colors_disabled
        time_str = '{:.3f}'.format(frame.time())

        if self.color:
            time_str = self._ansi_color_for_frame(frame) + time_str + colors.end

        result = u'{indent}{time_str} {function}  {c.faint}{code_position}{c.end}\n'.format(
            indent=indent,
            time_str=time_str,
            function=frame.function,
            code_position=frame.code_position_short,
            c=colors)

        children = [f for f in frame.children if f.proportion_of_total > 0.01]

        if children:
            last_child = children[-1]

        for child in children:
            if child is not last_child:
                c_indent = child_indent + (u'├─ ' if self.unicode else '|- ')
                cc_indent = child_indent + (u'│  ' if self.unicode else '|  ')
            else:
                c_indent = child_indent + (u'└─ ' if self.unicode else '`- ')
                cc_indent = child_indent + u'   '
            result += self.render(child, indent=c_indent, child_indent=cc_indent)

        return result

    def _ansi_color_for_frame(self, frame):
        colors = colors_enabled
        if frame.proportion_of_total > 0.6:
            return colors.red
        elif frame.proportion_of_total > 0.2:
            return colors.yellow
        elif frame.proportion_of_total > 0.05:
            return colors.green
        else:
            return colors.bright_green + colors.faint


class HTMLRenderer(Renderer):
    preferred_recorder = 'time_aggregating'

    def render(self, frame):
        resources_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'resources/')

        with open(os.path.join(resources_dir, 'style.css')) as f:
            css = f.read()

        with open(os.path.join(resources_dir, 'profile.js')) as f:
            js = f.read()

        with open(os.path.join(resources_dir, 'jquery-1.11.0.min.js')) as f:
            jquery_js = f.read()

        body = self.render_frame(frame)

        page = '''
            <html>
            <head>
                <style>{css}</style>
                <script>{jquery_js}</script>
            </head>
            <body>
                {body}
                <script>{js}</script>
            </body>
            </html>'''.format(css=css, js=js, jquery_js=jquery_js, body=body)

        return page

    def render_frame(self, frame):
        start_collapsed = all(child.proportion_of_total < 0.1 for child in frame.children)

        extra_class = ''
        extra_class += 'collapse ' if start_collapsed else ''
        extra_class += 'no_children ' if not frame.children else ''
        extra_class += 'application ' if frame.is_application_code else ''

        result = '''<div class="frame {extra_class}" data-time="{time}" date-parent-time="{parent_proportion}">
            <div class="frame-info">
                <span class="time">{time:.3f}s</span>
                <span class="total-percent">{total_proportion:.1%}</span>
                <!--<span class="parent-percent">{parent_proportion:.1%}</span>-->
                <span class="function">{function}</span>
                <span class="code-position">{code_position}</span>
            </div>'''.format(
                time=frame.time(),
                function=frame.function,
                code_position=frame.code_position_short,
                parent_proportion=frame.proportion_of_parent,
                total_proportion=frame.proportion_of_total,
                extra_class=extra_class)

        result += '<div class="frame-children">'

        # add this filter to prevent the output file getting too large
        children = [f for f in frame.children if f.proportion_of_total > 0.005]

        for child in children:
            result += self.render_frame(child)

        result += '</div></div>'

        return result


class colors_enabled:
    red = '\033[31m'
    green = '\033[32m'
    yellow = '\033[33m'
    blue = '\033[34m'
    cyan = '\033[36m'
    bright_green = '\033[92m'

    bold = '\033[1m'
    faint = '\033[2m'

    end = '\033[0m'


class colors_disabled:
    def __getattr__(self, key):
        return ''

colors_disabled = colors_disabled()



from __future__ import print_function
import time

from pyinstrument import Profiler

# Utilities #

def do_nothing():
    pass

def busy_wait(duration):
    end_time = time.time() + duration

    while time.time() < end_time:
        do_nothing()

def long_function_a():
    time.sleep(0.25)

def long_function_b():
    time.sleep(0.5)

# Tests #

def test_aggregator_collapses_multiple_calls():
    profiler = Profiler()
    profiler.start()

    long_function_a()
    long_function_b()
    long_function_a()
    long_function_b()

    profiler.stop()

    print(profiler.output_text())

    frame = profiler.first_interesting_frame()
    assert frame.function == 'test_aggregator_collapses_multiple_calls'
    assert len(frame.children) == 2

def test_timeline_retains_multiple_calls():
    profiler = Profiler(recorder='timeline')
    profiler.start()

    long_function_a()
    long_function_b()
    long_function_a()
    long_function_b()

    profiler.stop()

    print(profiler.output_text())

    frame = profiler.first_interesting_frame()
    assert frame.function == 'test_timeline_retains_multiple_calls'
    assert len(frame.children) == 4

def test_two_functions():
    profiler = Profiler()
    profiler.start()

    long_function_a()
    long_function_b()

    profiler.stop()

    print(profiler.output_text())

    frame = profiler.first_interesting_frame()

    assert frame.function == 'test_two_functions'
    assert len(frame.children) == 2

    frame_b, frame_a = frame.children

    assert frame_a.function == 'long_function_a'
    assert frame_b.function == 'long_function_b'
    assert 0.2 < frame_a.time() < 0.3
    assert 0.45 < frame_b.time() < 0.55



import sys, os
from operator import methodcaller


class Frame(object):
    """
    Object that represents a stack frame in the parsed tree
    """
    def __init__(self, identifier='', parent=None):
        self.identifier = identifier
        self.parent = parent
        self.self_time = 0

    @property
    def function(self):
        if self.identifier:
            return self.identifier.split('\x00')[0]

    @property
    def file_path(self):
        if self.identifier:
            return self.identifier.split('\x00')[1]

    @property
    def line_no(self):
        if self.identifier:
            return int(self.identifier.split('\x00')[2])

    @property
    def file_path_short(self):
        """ Return the path resolved against the closest entry in sys.path """
        if not hasattr(self, '_file_path_short'):
            if self.file_path:
                result = None

                for path in sys.path:
                    # On Windows, if self.file_path and path are on different drives, relpath
                    # will result in exception, because it cannot compute a relpath in this case.
                    # The root cause is that on Windows, there is no root dir like '/' on Linux.
                    try:
                        candidate = os.path.relpath(self.file_path, path)
                    except ValueError:
                        continue

                    if not result or (len(candidate.split(os.sep)) < len(result.split(os.sep))):
                        result = candidate

                self._file_path_short = result
            else:
                self._file_path_short = None

        return self._file_path_short

    @property
    def is_application_code(self):
        if self.identifier:
            return ('%slib%s' % (os.sep, os.sep)) not in self.file_path

    @property
    def code_position_short(self):
        if self.identifier:
            return '%s:%i' % (self.file_path_short, self.line_no)

    # stylistically I'd rather this was a property, but using @property appears to use twice
    # as many stack frames, so I'm forced into using a function since this method is recursive
    # down the call tree.
    def time(self):
        if not hasattr(self, '_time'):
            # can't use a sum(<generator>) expression here sadly, because this method
            # recurses down the call tree, and the generator uses an extra stack frame,
            # meaning we hit the stack limit when the profiled code is 500 frames deep.
            self._time = self.self_time

            for child in self.children:
                self._time += child.time()

        return self._time

    @property
    def proportion_of_parent(self):
        if not hasattr(self, '_proportion_of_parent'):
            if self.parent and self.time():
                try:
                    self._proportion_of_parent = self.time() / self.parent.time()
                except ZeroDivisionError:
                    self._proportion_of_parent = float('nan')
            else:
                self._proportion_of_parent = 1.0

        return self._proportion_of_parent

    @property
    def proportion_of_total(self):
        if not hasattr(self, '_proportion_of_total'):
            if not self.parent:
                self._proportion_of_total = 1.0
            else:
                self._proportion_of_total = self.parent.proportion_of_total * self.proportion_of_parent

        return self._proportion_of_total

    def add_child(self, child):
        self.children.append(child)

    @property
    def children(self):
        raise NotImplementedError()

    def __repr__(self):
        return 'Frame(identifier=%s, time=%f, len(children)=%d)' % (self.identifier, self.time(), len(self.children))


class TimelineFrame(Frame):
    def __init__(self, *args, **kwargs):
        self._children = []
        super(TimelineFrame, self).__init__(*args, **kwargs)

    def add_child(self, child):
        self._children.append(child)

    @property
    def children(self):
        return self._children


class TimeAggregatingFrame(Frame):
    def __init__(self, *args, **kwargs):
        self.children_dict = {}
        super(TimeAggregatingFrame, self).__init__(*args, **kwargs)

    def add_child(self, child):
        self.children_dict[child.identifier] = child

    @property
    def children(self):
        if not hasattr(self, '_children'):
            unsorted_children = self.children_dict.values()
            self._children = sorted(unsorted_children, key=methodcaller('time'), reverse=True)

        return self._children



# -*- coding: utf-8 -*-
import timeit
import warnings

from . import recorders
from . import renderers
from .util import object_with_import_path
from pyinstrument_cext import setstatprofile

timer = timeit.default_timer


class NotMainThreadError(Exception):
    '''deprecated as of 0.14'''
    pass


class SignalUnavailableError(Exception):
    '''deprecated as of 0.14'''
    pass


class Profiler(object):
    def __init__(self, use_signal=None, recorder='time_aggregating'):
        if use_signal is not None:
            warnings.warn('use_signal is deprecated and should no longer be used.', DeprecationWarning, stacklevel=2)

        self.interval = 0.001
        self.last_profile_time = 0

        self.recorder = get_recorder_class(recorder)()

    def start(self):
        self.last_profile_time = timer()
        setstatprofile(self._profile, self.interval)

    def stop(self):
        setstatprofile(None)

    def _profile(self, frame, event, arg):
        now = timer()
        time_since_last_signal = now - self.last_profile_time

        if time_since_last_signal < self.interval:
            return

        if event == 'call':
            frame = frame.f_back

        self.recorder.record_frame(frame, time_since_last_signal)

        self.last_profile_time = now

    def root_frame(self):
        """
        Returns the parsed results in the form of a tree of Frame objects
        """
        if not hasattr(self, '_root_frame'):
            self._root_frame = self.recorder.root_frame()

        return self._root_frame

    def first_interesting_frame(self):
        """
        Traverse down the frame hierarchy until a frame is found with more than one child
        """
        frame = self.root_frame()

        while len(frame.children) <= 1:
            if frame.children:
                frame = frame.children[0]
            else:
                # there are no branches
                return self.root_frame()

        return frame

    def starting_frame(self, root=False):
        if root:
            return self.root_frame()
        else:
            return self.first_interesting_frame()

    def output_text(self, root=False, unicode=False, color=False):
        return self.output(renderer='text', root=root, unicode=unicode, color=color)

    def output_html(self, root=False):
        return self.output(renderer='html', root=root)

    def output(self, renderer, root=False, **renderer_kwargs):
        renderer_class = get_renderer_class(renderer)
        renderer = renderer_class(**renderer_kwargs)
        return renderer.render(self.starting_frame(root=root))


def get_recorder_class(name):
    if name == 'time_aggregating':
        return recorders.TimeAggregatingRecorder
    elif name == 'timeline':
        return recorders.TimelineRecorder
    else:
        return object_with_import_path(name)


def get_renderer_class(name):
    if name == 'text':
        return renderers.ConsoleRenderer
    elif name == 'html':
        return renderers.HTMLRenderer
    else:
        return object_with_import_path(name)
